#!/usr/bin/env python3
"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.


A command line interface for Egeria Users - package info

This is an emerging capability based on the **click** package. Feedback welcome!

"""

from .ops_config import Config